# -*- coding: utf8 -*-
from __future__ import unicode_literals

import datetime
from collections import OrderedDict

from driverng import Worklist
from driverng.astm import (
    AstmMessageV2,
    HeaderSegment,
    HprimDate,
    HprimForeign,
    HprimList,
    HprimMultiple,
    HprimString,
    Segment,
)


class AcltopSegmentH(HeaderSegment):

    message_id = HprimString(3)
    emetteur = HprimString(5)
    recepteur = HprimString(10)
    # P: Production
    id_traitement = HprimString(12, default="P")
    version = HprimString(13, default="1394-97")
    date = HprimDate(14)

    patients = HprimForeign("P", "header")
    queries = HprimForeign("Q", "header")
    end = HprimForeign("L", "header", one2one=True, required=True)


class AcltopSegmentP(Segment):

    no_patient = HprimString(4)
    nom = HprimString((6, 1), max_length=20, trunc=True)
    prenom = HprimString((6, 2), max_length=20, trunc=True)
    date_naissance = HprimDate(8, fmt_date="%Y%m%d")

    # M/F/U
    sex = HprimString(9)
    code_correspondant = HprimString((14, 1))
    libelle_correspondant = HprimString((14, 2))

    orders = HprimForeign("O", "patient")


class AcltopSegmentO(Segment):

    tube = HprimString(3)
    instrument_id = HprimString(4)
    # /!\ Un seul test par segment O
    test = HprimString((5, 4))

    # Priorite :
    #  S : Stat Urgence
    #  R : Normal Routine
    urgence = HprimString(6)
    date_prelevement = HprimDate(7)

    # action code
    # A: Added
    # C: Cancel
    # Q: QC
    code_action = HprimString(12)

    # P: Plasma
    specimen_descriptor = HprimString(16)

    # Q: Response to query
    report_type = HprimList(26)

    results = HprimForeign("R", "order")


class AcltopSegmentR(Segment):

    test = HprimString((3, 4))
    valeur = HprimString(4)
    unite = HprimString(5)

    # Result Abnormal Flags
    #  L : Below low normal
    #  H : Above high normal
    #  N : Normal
    #  < : Below absolute low
    #  > : Above absolute high
    flag = HprimString(7)

    # Result status
    #  F : Final result
    #  P : Preliminary result
    #  V : result validated
    #  X : result cannot be done
    status = HprimList(9)
    operator_id = HprimString((11, 1))
    operator_passwd = HprimString((11, 2))

    # date et heure du resultat complete
    date = HprimDate(13)
    instrument_name = HprimString((14, 1))
    rack_id = HprimString((14, 2))
    position_no = HprimString((14, 3))


class AcltopSegmentC(Segment):

    commentaire = HprimString(3)

    # numero du commentaire
    no_code = HprimString((4, 1))
    # libelle du commentaire
    libelle = HprimString((4, 2))
    # code du commentaire
    code = HprimString((4, 3), max_length=2)
    # type du commentaire
    #  I : instrument flag comment
    #  N : negative comment
    type_commentaire = HprimString(5)


class AcltopSegmentQ(Segment):

    # Starting Range ID Number
    # Multiple numero d'echantillon d'un rack
    # range_id = HprimString(3)
    range_id = HprimList(3, basefield=HprimMultiple(0, tube=HprimString(2)))
    # request information status codes
    #  O : Request test order and demogs
    #  N : request new test orders still not downloaded
    request = HprimString(13)


class AcltopSegmentL(Segment):
    # code terminaison
    code = HprimString(3, default="N")


class Driver(AstmMessageV2):
    SPEC = {"code": "acltop", "label": "Acltop"}
    SEGMENT_CLASS_MAP = {
        "H": AcltopSegmentH,
        "P": AcltopSegmentP,
        "O": AcltopSegmentO,
        "R": AcltopSegmentR,
        "C": AcltopSegmentC,
        "Q": AcltopSegmentQ,
        "L": AcltopSegmentL,
    }
    DEFAULT_SEPARATORS = "|\\^&"

    OPTIONS = [
        AstmMessageV2.Option("code_emetteur", label="Code emetteur", default="LIS"),
        AstmMessageV2.Option(
            "code_recepteur", label="Code recepteur", default="ACL TOP"
        ),
        AstmMessageV2.Option("qc_prefix", label="Préfixe QC"),
        AstmMessageV2.BooleanOption(
            "qc_new_comportement",
            label="QC nouveau comportement: le nom du lot est O.3 et la référence est le champ C.4.2",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "qc_5653_prefix", label="Préfixer les QC 5653 par 'QUALIF'", default=True
        ),
        AstmMessageV2.BooleanOption(
            "proto_version_6",
            label="Version du protocole >= 6 (version transmettant la référence, le nom et niveau du lot de QC)",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "qc_reference_concatenee",
            label="Référence de lot QC: concatener O.3 avec C.4.2",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "qc_inclusion_alarmes",
            label="Intégrer les flags alarmes dans les résultats de QC",
            default=False,
        ),
        AstmMessageV2.BooleanOption(
            "acltop_preliminary_results",
            label="Récupérer les résultats préliminaires (R.9 = P)",
            default=False,
        ),
    ]

    def __init__(
        self,
        code_emetteur="LIS",
        code_recepteur="ACL TOP",
        qc_prefix=None,
        qc_new_comportement=False,
        qc_5653_prefix=True,
        proto_version_6=False,
        qc_reference_concatenee=False,
        qc_inclusion_alarmes=False,
        **kwargs
    ):
        self.code_emetteur = code_emetteur
        self.code_recepteur = code_recepteur
        self.qc_prefix = qc_prefix
        self.qc_new_comportement = qc_new_comportement
        self.qc_5653_prefix = qc_5653_prefix
        self.proto_version_6 = proto_version_6
        self.qc_reference_concatenee = qc_reference_concatenee
        self.qc_inclusion_alarmes = qc_inclusion_alarmes
        self.acltop_preliminary_results = kwargs.pop(
            "acltop_preliminary_results", False
        )
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        message = AcltopSegmentH(type_segment="H", **kwargs)
        message.end = AcltopSegmentL(type_segment="L")
        return message

    def encode_worklist(self, worklist, context):
        message = self.create(
            emetteur=self.code_emetteur,
            recepteur=self.code_recepteur,
            date=context["now"],
        )
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            action={"del": "C", worklist.ELSE: "A"},
            priority={"S": "S", "A": "S", worklist.ELSE: "R"},
        ):
            seg_p = message.patients.add(
                no_patient=patient.patient_id,
                nom=patient.last_name,
                prenom=patient.first_name,
                date_naissance=patient.birthdate,
                sex=patient.sex,
                code_correspondant=patient.location,
            )
            # Un segment O par test
            for sample, action, priority in patient.samples:
                for order in sample.orders:
                    if context["query"]:
                        report_type = ["Q"]
                    else:
                        report_type = ["O"]
                    seg_p.orders.add(
                        tube=sample.sample_id,
                        urgence=priority,
                        date_prelevement=sample.collection_date,
                        code_action=action,
                        report_type=report_type,
                        specimen_descriptor="P",
                        test=order.code,
                    )
        if message.patients:
            return message
        elif context["query"]:
            message.end.code = "I"
            return message

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for query in message.queries:
                for tube in query.range_id:
                    worklist.get_or_create(type="patient", sample_id=tube.tube)
            return worklist

        worklist = Worklist(context="result")
        for seg_p in message.patients:
            for seg_o in seg_p.orders:
                result_control = seg_o.code_action == "Q" or (
                    self.qc_prefix and seg_o.tube.startswith(self.qc_prefix)
                )
                lots_reactifs = OrderedDict()
                for seg_r in seg_o.results:
                    preliminary_results = False
                    if seg_r.status[0] == "P" and self.acltop_preliminary_results:
                        preliminary_results = True
                    if seg_r.status[0] not in ("F", "X") and not preliminary_results:
                        continue

                    alarmes = set()
                    lot = seg_o.tube
                    control_name = None
                    control_level = None
                    nom_reactif_concatene = True
                    for seg_c in seg_r.commentaires:

                        # control specific treatment
                        if result_control:
                            if (
                                self.qc_reference_concatenee
                                and lot in seg_c.no_code
                                and seg_c.libelle
                            ):
                                lot = "{}-{}".format(seg_o.tube, seg_c.libelle)
                            elif (
                                self.qc_reference_concatenee
                                and lot not in seg_c.no_code
                                and seg_c.libelle
                            ):
                                try:
                                    date_expiration = datetime.datetime.strptime(
                                        seg_c.code, "%Y-%m"
                                    )
                                except ValueError:
                                    continue
                                lots_reactifs[seg_c.libelle] = {
                                    "lot": seg_c.libelle,
                                    "expiration": date_expiration,
                                    "type": "control",
                                }
                            elif (
                                self.qc_new_comportement
                                and lot in seg_c.no_code
                                and seg_c.libelle
                            ):
                                lot = seg_c.libelle
                            elif (
                                self.qc_new_comportement
                                and lot not in seg_c.no_code
                                and seg_c.libelle
                            ):
                                lots_reactifs[seg_c.libelle] = {
                                    "lot": seg_c.libelle,
                                    "expiration": datetime.datetime.strptime(
                                        seg_c.code, "%Y-%m"
                                    ),
                                    "type": "control",
                                }
                            elif not self.proto_version_6 and lot in seg_c.no_code:
                                control_level = seg_c.no_code[3:]
                                lot = seg_c.libelle
                                nom_reactif_concatene = False
                            if (
                                self.qc_5653_prefix
                                and seg_c.no_code == "5653"
                                and seg_c.code == "ME"
                            ):
                                lot = "QUALIF-{}".format(seg_o.tube)
                            if self.proto_version_6 and lot in seg_c.no_code:
                                control_level = seg_c.libelle
                                control_name = seg_c.no_code[3:]
                            elif (
                                self.proto_version_6
                                and "LI" in seg_c.no_code
                                and lot not in seg_c.no_code
                            ):
                                lots_reactifs[seg_c.libelle] = {
                                    "lot": seg_c.libelle,
                                    "expiration": datetime.datetime.strptime(
                                        seg_c.code, "%Y-%m"
                                    ),
                                    "type": "control",
                                }

                        # Lots de réactifs
                        if seg_c.no_code.startswith("LI-"):
                            if nom_reactif_concatene:
                                ref_lot = "{}{}".format(
                                    seg_c.no_code[3:], seg_c.libelle
                                )
                            else:
                                ref_lot = "{}".format(seg_c.libelle)
                            date = None
                            if seg_c.code:
                                date = datetime.datetime.strptime(seg_c.code, "%Y-%m")
                            lots_reactifs[seg_c.no_code[3:]] = {
                                "lot": ref_lot,
                                "expiration": date,
                                "type": "reagent",
                            }
                        elif seg_c.type_commentaire == "I":
                            alarmes.add(seg_c.no_code)
                            alarmes.add(seg_c.code)

                    # le code resultat est la concatenation
                    # du test avec l'unite s'il y en a une
                    if seg_r.unite:
                        result_code = "{}{}".format(seg_r.test, seg_r.unite)
                    else:
                        result_code = seg_r.test

                    result_data = dict()
                    rack = "{}-{}".format(seg_r.rack_id, seg_r.position_no)
                    if result_control:
                        sample = worklist.get_or_create(
                            type="control", sample_id=lot or ""
                        )
                        sample.control_lot = lot
                        sample.control_name = control_name
                        sample.control_level = control_level
                    else:
                        sample = worklist.get_or_create(
                            type="patient", sample_id=seg_o.tube, rack=rack
                        )
                    if lots_reactifs:
                        result_data["lots"] = list(lots_reactifs.values())
                    if seg_r.flag:
                        alarmes.add(seg_r.flag)
                    if (
                        not result_control
                        or result_control
                        and self.qc_inclusion_alarmes
                    ):
                        result_data["flags"] = sorted(alarmes)
                    if not result_control and seg_r.operator_id:
                        result_data["operator"] = seg_r.operator_id
                    sample.results.add(
                        code=result_code,
                        value=seg_r.valeur,
                        date=seg_r.date,
                        **result_data
                    )
        return worklist
